///////////////////////////////////////////////////////////////////////////////
//
//      ##  ######
//       ######  ###
//  ## ###############        Shark 3D Engine (www.shark3d.com)
//   ########## # # #
//    ########                Copyright (c) 1996-2006 Spinor GmbH.
//   ######### # # #          All rights reserved.
//  ##   ##########
//      ##
//
///////////////////////////////////////////////////////////////////////////////

//@cpp

#ifndef GTP_SHADER_CAUSTICCUBE_H
#define GTP_SHADER_CAUSTICCUBE_H

#include "../../interf/eng_gfxvar.h"
#include "../../interf/eng_shader.h"
#include "../../interf/eng_shaderenv.h"
#include "../../util/eng_util_shaderdest.h"
#include "../../util/eng_util_shaderexec.h"
#include "../../../comp/core/comp_setup.h"
#include "../../../comp/util/comp_util_uniquebase.h"
#include "../../../util/mempool/util_mempool_array_types.h"

///////////////////////////////////////////////////////////////////////////////

struct gtp_Shader_CausticCubeDestGfx
{
    s3d_CDrvVarSlot m_FromVarSlot;
    s3d_CDrvVarSlot m_ToVarSlot;
    int m_TexProp;
};

typedef s3d_CUtilArray<gtp_Shader_CausticCubeDestGfx>
        gtp_Shader_CausticCubeDestGfxArray;

typedef const gtp_Shader_CausticCubeDestGfxArray &
        gtp_Shader_CausticCubeDestGfxArray_cr;

///////////////////////////////////////////////////////////////////////////////

class gtp_Shader_CausticCubeGfx;

//@
typedef s3d_CUtilKeyEntry<s3d_CEngGfxCtxPtr, gtp_Shader_CausticCubeGfx*>
        gtp_Shader_CausticCubeGfxEntry;

//@
typedef s3d_CUtilTree<gtp_Shader_CausticCubeGfxEntry>
        gtp_Shader_CausticCubeGfxTree;

///////////////////////////////////////////////////////////////////////////////

struct gtp_Shader_CausticCubeData: s3d_CUtilPtrBase
{
    s3d_CUtilStr m_Info;
    s3d_CUtilAtomArray m_MethodArray;
    s3d_CUtilStr m_RankInfo;
    int m_Rank;
    int m_IterStart;
    int m_IterEnd;
    int m_IterStep;
    float m_IterWidthScale;
    float m_IterHeightScale;
    int m_DestProp;
    float m_WidthScale;
    float m_HeightScale;
    s3d_CUtilStr m_TaskInfoVarInfo;
    s3d_CUtilAtomPtr m_TaskInfoVarAtom;
    s3d_CUtilStr m_SrcVarInfo;
    s3d_CUtilAtomArray m_SrcVarAtomArray;
    s3d_CEngUtilShaderDest m_Dest;
    s3d_CUtilStr m_ProgInfo;
    s3d_CUtilStr m_ProgIdent;
    s3d_CUtilStr m_ProgParamInfo;
    s3d_CUtilAtomPairArray m_ProgParamMapArray;

	int m_UpdateInterval;
	int m_StartFrame;
	bool m_UpdateAllFace;
	bool m_UsePoints;
	int m_PhotonCount;
};

typedef s3d_CUtilPtr<gtp_Shader_CausticCubeData>
        gtp_Shader_CausticCubeDataPtr;

///////////////////////////////////////////////////////////////////////////////

//@
class gtp_Shader_CausticCube
    : public s3d_CEngShader,
    public s3d_CCompUtilUniqueBase,
    public s3d_CCompSetup
{
public:
    S3D_UTIL_RTTI_TABLE_DECLARE

    //@
    gtp_Shader_CausticCube();

    //@
    void SetupInit(
            s3d_CUtilMsgHandler *MsgHandler, s3d_CUtilStr_cr Info,
            s3d_CCompSuppl *FetchSuppl, s3d_CUtilSnkChunk *Param);

    //@
    void SetupDone();

    //@
    s3d_CUtilStr GetInfo();

    //@
    void RegisterNotif(s3d_CUtilNotifRecip *Recip);

    //@
    s3d_CEngShaderGfxPtr FindShaderGfx(
            s3d_CUtilNotifGather *NotifGather, s3d_CEngGfxCtx *GfxCtx);

private:
    s3d_CEngShaderEnvPtr m_ShaderEnv;
    gtp_Shader_CausticCubeGfxTree m_ShaderGfxTree;
    gtp_Shader_CausticCubeDataPtr m_Data;

    friend class gtp_Shader_CausticCubeGfx;
    friend class gtp_Shader_CausticCubeInst;
};

//@
typedef s3d_CUtilPtr<gtp_Shader_CausticCube>
        gtp_Shader_CausticCubePtr;

///////////////////////////////////////////////////////////////////////////////

//@
class gtp_Shader_CausticCubeGfx: public s3d_CEngShaderGfx
{
public:
    //@
    gtp_Shader_CausticCubeGfx(
            s3d_CEngShaderEnv *ShaderEnv, 
            s3d_CUtilNotifGather *NotifGather,
            s3d_CEngGfxCtx *GfxCtx, 
            gtp_Shader_CausticCubeData *Data);

    //@
    gtp_Shader_CausticCubeGfxTree::CNode *GetNode();

    //@
    void RegisterNotif(s3d_CUtilNotifRecip *Recip);

    //@
    void ExtractGfx();

    //@
    bool IsValid();

    //@
    void LinkCollect(s3d_CUtilAtomSet &VarAtomSet);

    //@
    void LinkAppoint(s3d_CDrvVarDecl *VarDecl);

    //@
    s3d_CEngShaderInstPtr CreateShaderInst(
            s3d_CEngShaderHead *Head);

private:
    gtp_Shader_CausticCubeGfxTree::CNode m_Node;
    s3d_CEngShaderEnvPtr m_ShaderEnv;
    gtp_Shader_CausticCubeDataPtr m_Data;

    // ShaderGfx-Data Variables:
    s3d_CEngGfxProgPtr m_GfxProg;
    s3d_CDrvVarSlotArray m_SrcVarSlotArray;
    s3d_CEngUtilShaderDestGfx m_FilterDestGfx;
    s3d_CDrvVarSlotInit m_TaskInfoVarSlot;
    s3d_CDrvVarDeclPtr m_ProgParamDecl;

    friend class gtp_Shader_CausticCubeInst;
};

//@
typedef s3d_CUtilPtr<gtp_Shader_CausticCubeGfx> 
        gtp_Shader_CausticCubeGfxPtr;

///////////////////////////////////////////////////////////////////////////////

//@

class gtp_Shader_CausticCubeInst
    : public s3d_CEngShaderInst,
    public s3d_CEngShaderExec
{
public:
    //@
    gtp_Shader_CausticCubeInst(
            s3d_CEngShaderEnv *ShaderEnv,
            s3d_CEngGfxCtx *GfxCtx,
            gtp_Shader_CausticCubeGfx *ShaderGfx,
            s3d_CEngGfxProg *GfxProg);

    //@
    void RegisterNotif(s3d_CUtilNotifRecip *Recip);

    //@
    void SetShaderCtx(s3d_CEngShaderCtx *ShaderCtx);

    //@
    void SetGfxState(s3d_CEngGfxState *GfxState);

    //@
    void SetPos(s3d_CEngPos *Pos);

    //@
    void SetModel(s3d_CEngModel *Model);

    //@
    void ArrangeCollect(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray &ScopeArray);

    //@
    void ArrangeInit(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray_cr ScopeArray);

    //@
    void ArrangePerform(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray_cr ScopeArray);

    //@
    void ArrangeFeedback(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray_cr ScopeArray);

    //@
    void ArrangeGather(
            s3d_CEngShaderExecArray &ExecArray);

    //@
    void ShaderExecPerform(
            s3d_CEngShaderScope *Scope, 
            s3d_CUtilAtom *Method,
            s3d_CEngGfxTaskArray &TaskArray);

    //@
    void *GfxElemJobExec(
            s3d_CEngShaderScope *Scope,
            void *GfxMark);

private:
    int m_FrameCount;
    int m_CurrentFace;	
   
	s3d_CEngShaderEnvPtr m_ShaderEnv;
    s3d_CEngGfxCtxPtr m_GfxCtx;
    gtp_Shader_CausticCubeGfxPtr m_ShaderGfx;
    s3d_CEngGfxProgPtr m_GfxProg;

    void Iteration(
            s3d_CEngShaderScope *Scope, int Iter,
            int DestProp, float WantWidth, float WantHeight);

	
    void FetchSrcArray(
            s3d_CUtilMemPool *MemPool, s3d_CEngShaderScope *Scope, 
            const s3d_CSysChar *Info,
            s3d_CEngGfxVarTexChanArray &FilterSrcTexArray,
            s3d_CDrvVarSlotArray &SrcVarSlotArray);

    void CalcSrcTexSize(
            s3d_CUtilMemPool *MemPool, s3d_CEngShaderScope *Scope,
            s3d_CDrvGfxEng *GfxEng, int &Width, int &Height,
            s3d_CEngGfxVarTexChanArray &FilterSrcTexArray);
};

///////////////////////////////////////////////////////////////////////////////

#endif
