///////////////////////////////////////////////////////////////////////////////
//
//      ##  ######
//       ######  ###
//  ## ###############        Shark 3D Engine (www.shark3d.com)
//   ########## # # #
//    ########                Copyright (c) 1996-2006 Spinor GmbH.
//   ######### # # #          All rights reserved.
//  ##   ##########
//      ##
//
///////////////////////////////////////////////////////////////////////////////

//@cpp

#ifndef GTP_SHADER_ENVMAP_H
#define GTP_SHADER_ENVMAP_H

#include "../../interf/eng_shader.h"
#include "../../interf/eng_shaderenv.h"
#include "../../util/eng_util_shaderdest.h"
#include "../../../comp/core/comp_setup.h"
#include "../../../comp/util/comp_util_uniquebase.h"

///////////////////////////////////////////////////////////////////////////////

class gtp_Shader_EnvMapGfx;

//@
typedef s3d_CUtilKeyEntry<s3d_CEngGfxCtxPtr, gtp_Shader_EnvMapGfx*>
        gtp_Shader_EnvMapGfxEntry;

//@
typedef s3d_CUtilTree<gtp_Shader_EnvMapGfxEntry>
        gtp_Shader_EnvMapGfxTree;

///////////////////////////////////////////////////////////////////////////////

struct gtp_Shader_EnvMapData: s3d_CUtilPtrBase
{
    s3d_CUtilStr m_Info;

    s3d_CUtilStr m_MeshVarInfo;
    s3d_CUtilAtomPtr m_MeshVarAtom;
    s3d_CUtilStr m_TexChanInfo;
    int m_DestProp;
    s3d_CEngUtilShaderDest m_Dest;

    float m_NegZ;
    float m_PosZ;
    int m_MaxExt;
    float m_ExtDistScale;
    int m_MaxRecursion;

	int m_UpdateInterval;
	int m_StartFrame;
	bool m_UpdateAllFace;
	bool m_WorldSpace;
	s3d_CUtilAtomPtr m_LastCenterVarAtom;
	s3d_CUtilStr m_LastCenterVarInfo;

    s3d_CUtilStr m_RankInfo;
    int m_Rank;
    s3d_CUtilAtomArray m_MethodArray;
    s3d_CUtilAtomPtr m_EnumTrigger;
};

typedef s3d_CUtilPtr<gtp_Shader_EnvMapData>
        gtp_Shader_EnvMapDataPtr;

///////////////////////////////////////////////////////////////////////////////

//@
class gtp_Shader_EnvMap
    : public s3d_CEngShader,
    public s3d_CCompUtilUniqueBase,
    public s3d_CCompSetup
{
public:
    S3D_UTIL_RTTI_TABLE_DECLARE

    //@
    gtp_Shader_EnvMap();

    //@
    void SetupInit(
            s3d_CUtilMsgHandler *MsgHandler, s3d_CUtilStr_cr Info,
            s3d_CCompSuppl *FetchSuppl, s3d_CUtilSnkChunk *Param);

    //@
    void SetupDone();

    //@
    s3d_CUtilStr GetInfo();

    //@
    void RegisterNotif(s3d_CUtilNotifRecip *Recip);

    //@
    s3d_CEngShaderGfxPtr FindShaderGfx(
            s3d_CUtilNotifGather *NotifGather, s3d_CEngGfxCtx *GfxCtx);

private:
    s3d_CEngShaderEnvPtr m_ShaderEnv;
    gtp_Shader_EnvMapGfxTree m_ShaderGfxTree;
    gtp_Shader_EnvMapDataPtr m_Data;

    friend class gtp_Shader_EnvMapGfx;
    friend class gtp_Shader_EnvMapInst;
};

//@
typedef s3d_CUtilPtr<gtp_Shader_EnvMap>
        gtp_Shader_EnvMapPtr;

///////////////////////////////////////////////////////////////////////////////

//@
class gtp_Shader_EnvMapGfx: public s3d_CEngShaderGfx
{
public:
    //@
    gtp_Shader_EnvMapGfx(
            s3d_CEngShaderEnv *ShaderEnv,
            s3d_CEngGfxCtx *GfxCtx,
            gtp_Shader_EnvMapData *Data);

    //@
    gtp_Shader_EnvMapGfxTree::CNode *GetNode();

    //@
    void RegisterNotif(s3d_CUtilNotifRecip *Recip);

    //@
    void ExtractGfx();

    //@
    bool IsValid();

    //@
    void LinkCollect(s3d_CUtilAtomSet &VarAtomSet);

    //@
    void LinkAppoint(s3d_CDrvVarDecl *VarDecl);

    //@
    s3d_CEngShaderInstPtr CreateShaderInst(
            s3d_CEngShaderHead *Head);

private:
    gtp_Shader_EnvMapGfxTree::CNode m_Node;
    s3d_CEngShaderEnvPtr m_ShaderEnv;
    gtp_Shader_EnvMapDataPtr m_Data;

    // ShaderGfx-Data Variables:
    s3d_CDrvVarSlotInit m_MeshVarSlot;
    s3d_CEngUtilShaderDestGfx m_DestGfx;
	s3d_CDrvVarSlotInit m_LastCenterVarSlot;

    friend class gtp_Shader_EnvMapInst;
};

//@
typedef s3d_CUtilPtr<gtp_Shader_EnvMapGfx> 
        gtp_Shader_EnvMapGfxPtr;

///////////////////////////////////////////////////////////////////////////////

//@
class gtp_Shader_EnvMapInst
    : public s3d_CEngShaderInst,
    public s3d_CEngShaderExec

{
public:
    //@
    gtp_Shader_EnvMapInst(
            s3d_CEngShaderEnv *ShaderEnv,
            s3d_CEngGfxCtx *GfxCtx, s3d_CEngShaderHead *Head,
            gtp_Shader_EnvMapGfx *GfxShader);

    //@
    void RegisterNotif(s3d_CUtilNotifRecip *Recip);

    //@
    void SetShaderCtx(s3d_CEngShaderCtx *ShaderCtx);

    //@
    void SetGfxState(s3d_CEngGfxState *GfxState);

    //@
    void SetPos(s3d_CEngPos *Pos);

    //@
    void SetModel(s3d_CEngModel *Model);

    //@
    void ArrangeCollect(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray &ScopeArray);

    //@
    void ArrangeInit(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray_cr ScopeArray);

    //@
    void ArrangePerform(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray_cr ScopeArray);

    //@
    void ArrangeFeedback(
            s3d_CUtilMemPool *MemPool,
            s3d_CEngShaderScopeArray_cr ScopeArray);

    //@
    void ArrangeGather(
            s3d_CEngShaderExecArray &ExecArray);

    //@
    void ShaderExecPerform(
            s3d_CEngShaderScope *Scope, 
            s3d_CUtilAtom *Method,
            s3d_CEngGfxTaskArray &TaskArray);

    //@
    void *GfxElemJobExec(
            s3d_CEngShaderScope *Scope,
            void *GfxMark);

private:
	int m_FrameCount;
    int m_CurrentFace;
	s3d_CEngShaderEnvPtr m_ShaderEnv;
    gtp_Shader_EnvMapGfxPtr m_ShaderGfx;
    s3d_CEngGfxCtxPtr m_GfxCtx;
    s3d_CEngShaderHeadEye m_Head;
    bool m_CanRenderColorAlphaTex;
	s3d_CUtilVec4f m_LastCenter;

    void Calc(
            s3d_CUtilMemPool *MemPool, s3d_CEngShaderScope *Scope, 
            s3d_CUtilTranslQuatf_cr ViewTransf,
            s3d_CUtilVec3f_cr BoundCen, s3d_CUtilVec3f_cr BoundExt,
            s3d_CEngUtilShaderDestJob *&DestJob);
	
	void SetVars(
            s3d_CUtilMemPool *MemPool, s3d_CEngShaderScope *Scope);

    
};

//@
typedef s3d_CUtilPtr<gtp_Shader_EnvMapInst>
        gtp_Shader_EnvMapInstPtr;

///////////////////////////////////////////////////////////////////////////////

#endif
